import 'dart:io';
import 'dart:async';
import 'package:chopper/chopper.dart';
import 'package:adeptarch/src/api/adept_chopper_client.dart';
import 'package:adeptutils/src/exception_utils.dart';

BackendAPI backendAPI = BackendAPI();

class BackendAPI
{
  AdeptChopperClient chopperClient;
  Converter converter;
  final Map<Type, ChopperService> _services = {};

  Duration timeout;
  String apiProtocol;
  String apiHost;
  String apiPort;
  String apiUrl;

  String webProtocol;
  String webHost;
  String webPort;
  String webUrl;

  String apiDateFormat;
  bool trustAllCerts;

  BackendAPI(){
    apiProtocol = "http://";
    apiHost = "www.myapi.com";
    apiPort = "8080";
    apiUrl = apiProtocol + apiHost + apiPort ?? "" + "/api";
    apiDateFormat = "yyyy-MM-dd";

    webProtocol = "http://";
    webHost = "www.myweb.com";
    webPort = "";
    webUrl = webProtocol + webHost + webPort ?? "";

    trustAllCerts = true;
  }

  void init()
  {
    try{chopperClient.dispose();} catch(exception, stackTrace){}
    converter ??= JsonConverter();
    chopperClient = AdeptChopperClient(
      baseUrl: apiUrl,
      services: _services.values,
      converter: converter,
    );
    chopperClient.timeout = timeout;
    _initServices();
  }

  void _initServices(){
    _services.forEach((type, service) {
      service.client = chopperClient;
    });
  }

  void dispose(){
    try{
      _services.forEach((t, s) => s.dispose());
      _services.clear();
      chopperClient.dispose();
    }
    catch(exception, stackTrace){}
  }

  void addService(ChopperService service){
    if(service != null) {
      service.client = chopperClient;
      _services[service.definitionType] = service;
    }
  }

  T service<T extends ChopperService>(Type type) {
    final s = _services[type];
    if (s == null) {
      throw Exception("Service of type '$type' not found.");
    }
    return s;
  }

  void updateApiUrl({String apiUrl, String apiProtocol, String apiHost,
    String apiPort})
  {
    if(apiUrl == null || apiUrl.isEmpty){
      this.apiProtocol = apiProtocol;
      this.apiHost = apiHost;
      this.apiPort = apiPort;
      this.apiUrl = apiProtocol + apiHost + apiPort ?? "" + "/api/";
    } else{
      this.apiUrl = apiUrl;
    }
  }

  void updateWebUrl({String webUrl, String webProtocol, String webHost,
    String webPort})
  {
    if(webUrl == null || webUrl.isEmpty){
      this.webProtocol = webProtocol;
      this.webHost = webHost;
      this.webPort = webPort;
      this.webUrl = webProtocol + webHost + webPort ?? "";
    } else {
      this.webUrl = webUrl;
    }
  }
  
  ///Implement this method when checked what exceptions could be
  String getLocalizedMessageFromException(dynamic exception)
  {
    if(exception == null) return null;
    if(exception is GenericException) return getLocalizedMessageFromException((exception as GenericException).causeException);

    if(exception is WebSocketException || exception is TimeoutException ||
        exception.toString().toLowerCase().contains("TimeoutException"))
      return "Server connection timeout.";
    if(exception is SocketException || exception.toString().toLowerCase().contains("failed to connect"))
    return "Check your internet connection.";
    if(exception is HandshakeException)
      return exception.toString();

    return null;
  }

  String getGenericLocalizedMessageFromException(dynamic exception)
  {
    String message = getLocalizedMessageFromException(exception);
    return message ?? "Unknown error";
  }

  String getLocalizedMessageFromHttpResultCode(int code)
  {
    if(code == null) return null;
    switch(code)
    {
      case HttpStatus.noContent:
        return "($code) No Content";

      case HttpStatus.badRequest:
        return "($code) Bad Request";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized";

      case HttpStatus.forbidden:
        return "($code) Forbidden";

      case HttpStatus.notFound:
        return "($code) Not Found";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error";
    }
    return "HTTP error code: ($code)";
  }

//  OkHttpClient.Builder getOkHttpClientBuilder()
//  {
//    //Change the time values
//    OkHttpClient.Builder okHttpClientBuilder = trustAllCerts ? UnsafeOkHttpClient.getUnsafeOkHttpBuilder() : new OkHttpClient().newBuilder();
//    okHttpClientBuilder.readTimeout(25, TimeUnit.SECONDS);
//    okHttpClientBuilder.connectTimeout(25, TimeUnit.SECONDS);
//    okHttpClientBuilder.writeTimeout(25, TimeUnit.SECONDS);
//
////        okHttpClientBuilder.certificatePinner(new CertificatePinner.Builder().add("*.androidadvance.com", "sha256/RqzElicVPA6LkKm9HblOvNOUqWmD+4zNXcRb+WjcaAE=")
////            .add("*.xxxxxx.com", "sha256/8Rw90Ej3Ttt8RRkrg+WYDS9n7IS03bk5bjP/UXPtaY8=")
////            .add("*.xxxxxxx.com", "sha256/Ko8tivDrEjiY90yGasP6ZpBU4jwXvHqVvQI0GS3GNdA=")
////            .add("*.xxxxxxx.com", "sha256/VjLZe/p3W/PJnd6lL8JVNBCGQBZynFLdZSTIqcO0SJ8=")
////            .build());
////
////        int cacheSize = 10 * 1024 * 1024; // 10 MiB
////        Cache cache = new Cache(context.getCacheDir(), cacheSize);
////        okHttpClientBuilder.cache(cache);
//    return okHttpClientBuilder;
//  }
//
//  Retrofit singletonInstanceWithGsonConverterFactory()
//  {
//    return singletonInstanceWithConverterFactory(GsonConverterFactory.create());
//  }
//
//  Retrofit singletonInstanceWithConverterFactory(Converter.Factory converterFactory)
//  {
//    Instance tempInstance;
//
//    if(converterFactory instanceof GsonConverterFactory)
//      tempInstance = Instance.GSON_CONVERTER_FACTORY;
//    else
//      tempInstance = Instance.OTHER;
//
//    if(retrofit == null || instance != tempInstance || !retrofit.baseUrl().toString().equals(apiUrl))
//    {
//      instance = tempInstance;
//      retrofit = new Retrofit.Builder()
//          .client(getOkHttpClientBuilder().build())
//          .addConverterFactory(converterFactory)
//          .baseUrl(apiUrl)
//          .build();
//    }
//    return retrofit;
//  }
//
//  Retrofit newInstanceWithBaseUrl(String base_url)
//  {
//    return new Retrofit.Builder()
//        .client(getOkHttpClientBuilder().build())
//        .addConverterFactory(GsonConverterFactory.create())
//        .baseUrl(base_url)
//        .build();
//  }
}